/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2010, Sangoma Technologies Corp.
 *
 * Moises Silva <moy@sangoma.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief codec_sangoma.c - Sangoma Vocallo codec module
 * 
 * \ingroup codecs
 */

/*** MODULEINFO
	<depend>sangoma_transcode</depend>
 ***/

/*
 * some notes about Asterisk versions and its interfaces, versions may vary slightly, I tested
 * with latest of each release, feel free to adjust if you find something inaccurate
 * > 1602 (at this time trunk) defines new RTP pluggable architecture and some new config keywords
 * >= 10601 defines data as a union instead of a pointer inside an ast_frame (f.data.ptr)
 * >= 10600 defines new CLI macros and prototypes for the callbacks 
 * the asteris/version.h macro ASTERISK_VERSION_NUM is used to keep this source compatible with all versions
 * >= 10400 asterisk.h was introduced
 *    ast_frame now has flags member
 *    the transcoding interface changed a lot, mostly the way the private structure is allocated and used
 * */
#include "asterisk/version.h"

#if ASTERISK_VERSION_NUM >= 10400
#include "asterisk.h"
#endif

/* common Asterisk includes */
#include "asterisk/module.h"
#include "asterisk/config.h"
#include "asterisk/translate.h"
#include "asterisk/alaw.h"
#include "asterisk/utils.h"
#include "asterisk/sched.h"
#include "asterisk/cli.h"
#include "asterisk/acl.h"
#include "asterisk/app.h"

#if ASTERISK_VERSION_NUM > 10602
#include "asterisk/rtp_engine.h"
#else
#include "asterisk/rtp.h"
#endif

/* Sangoma transcoding library */
#ifdef SNGTC_DEBUG_HARD_LINK
#include <sng_tc/sng_tc.h>
#else
#include <sng_tc/sngtc_node.h>
#endif

#include <sys/ioctl.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <errno.h>

/* default to 1 until bug http://issues.asterisk.org/view.php?id=17092 is fixed in Asterisk */
static int g_attempt_unload_conflicting_translators = 1;

/* we've seen delays, crashes and audio drops related to non-DAHDI timing, do not allow other timings by default */
static int g_allow_nondahdi_timing = 0;
static int g_dahdi_timing = 0;

/* supported codecs */
typedef struct vocallo_codec_s {
	int codec_id;
	int astformat;
	int maxms;
} vocallo_codec_t;

vocallo_codec_t g_codec_map[] =
{
	{ SNGTC_CODEC_PCMU,    AST_FORMAT_ULAW    , 40 },
	{ SNGTC_CODEC_PCMA,    AST_FORMAT_ALAW    , 40 },
	{ SNGTC_CODEC_L16_1,   AST_FORMAT_SLINEAR , 40 },
	{ SNGTC_CODEC_G729AB,  AST_FORMAT_G729A   , 40 },
	{ SNGTC_CODEC_G722,    AST_FORMAT_G722    , 20 },
	{ SNGTC_CODEC_G726_32, AST_FORMAT_G726    , 40 }, 
	{ SNGTC_CODEC_GSM_FR,  AST_FORMAT_GSM     , 20 },
#if 0
	{ SNGTC_CODEC_ILBC,    AST_FORMAT_ILBC    , 0 },
#endif
	{ -1, -1 },
};

#define MAX_CODECS_LIST 64
#define MAX_CODEC_STR 32
static unsigned int g_codec_register_list_count = 0;
static unsigned int g_codec_noregister_list_count = 0;
static char g_codec_register_list[MAX_CODECS_LIST][MAX_CODEC_STR];
static char g_codec_noregister_list[MAX_CODECS_LIST][MAX_CODEC_STR];

#define SANGOMA_TRANSLATOR_BUFFER_SIZE 8000
#define FAKE_FRAME_IDLE 0
#define FAKE_FRAME_REQUESTED 1
#define FAKE_FRAME_PROVIDED 2
#define FAKE_FRAME_TERMINATED 3 

#define SANGOMA_DEFAULT_SAMPLES 80*3
#define SANGOMA_DEFAULT_SAMPLING_RATE 80
#define SANGOMA_DEFAULT_RTP_STACK  "asterisk" 
#define SANGOMA_DEFAULT_UDP_PORT 15000
#define SANGOMA_TRANSCODE_CONFIG "sangoma_codec.conf" 

#define SANG_SESS_WARN_TX(...) if (sess->txwarn > 10) { \
					sess->txwarn = sess->txwarn > 1000 ? 0 : sess->txwarn; \
					sess->txwarn++; \
				} else { \
					ast_log(LOG_WARNING, __VA_ARGS__); \
					sess->txwarn++; \
				}

#define SANG_SESS_WARN_RX(...) if (sess->rxwarn > 10) {  \
					sess->rxwarn = sess->rxwarn > 1000 ? 0 : sess->rxwarn; 		\
					sess->rxwarn++; \
				} else { \
					ast_log(LOG_WARNING, __VA_ARGS__); \
					sess->rxwarn++; \
				}

/*! transcoder init configuration */
static sngtc_init_cfg_t g_init_cfg;

enum sangoma_rtp_flags {
	SNG_RTP_TX = (1 << 0),
	SNG_RTP_RX = (1 << 1),
};

struct sangoma_rtp_handle {
	void *ast_rtp; /* The asterisk RTP handle */
	unsigned int flags;
	ast_mutex_t rxlock;
	ast_mutex_t txlock;
};

/* \brief sangoma transcoding session private structure */
#if ASTERISK_VERSION_NUM >= 10400
#define asterisk_transcoding_pvt ast_trans_pvt
#define sangoma_transcoding_session sangoma_transcoding_session
#define sangoma_get_pvt(pvt) (pvt)->pvt
#else
#define asterisk_transcoding_pvt ast_translator_pvt
#define sangoma_transcoding_session ast_translator_pvt
#define sangoma_get_pvt(pvt) (pvt)
#endif
struct sangoma_transcoding_session {
	struct asterisk_transcoding_pvt *owner;
	int fakestate;
	sngtc_codec_request_t codec_request;
	sngtc_codec_reply_t codec_reply;
#ifdef SNGTC_DEBUG_HARD_LINK
	sngtc_rtp_stats_t rtp_stats;
#endif

	/* tx and rx rtp streams */
	struct sangoma_rtp_handle *txrtp;
	struct sangoma_rtp_handle *rxrtp;

	/* unique session id */
	unsigned long long sessid;
	
	/* tx calls and rx calls counters */
	unsigned long long txcount;
	unsigned long long rxcount;
	unsigned long long rxcallcount;
	unsigned long long rxnullcount;

	unsigned short txavgtime;
	unsigned short rxavgtime;
	unsigned short rxcallavgtime;

	struct timeval lasttxtime;
	struct timeval lastrxtime;
	struct timeval lastrxcalltime;

	/* seq counters */
  	long lastrxseqno;
	long lostrxpackets;
	/* warning counters */
	unsigned short rxwarn;
	unsigned short txwarn;
#if ASTERISK_VERSION_NUM < 10400
	/* this duplicates some info like sangoma_transcoding_session which is the same as owner 
	 * but keeps the processing code across asterisk versions cleaner (less if ASTERISK_VERSION stuff) */
	int samples;
	int srcfmt;
	int dstfmt;
	struct sangoma_transcoding_session *t;
	struct ast_frame f;
#endif
	AST_LIST_ENTRY(sangoma_transcoding_session) entry;
};

/* \brief simple container to list all registered asterisk translators in this sangoma module */
struct sangoma_translator {
	struct ast_translator t;
	AST_LIST_ENTRY(sangoma_translator) entry;
};

/* \brief global scheduling context */
static struct sched_context *g_sched = NULL;

/* \brief global session id, the list of sangoma translator sessions lock protects the global next id as well */
static unsigned long long g_next_session_id = 0;
static unsigned long long g_total_sessions  = 0;

/* \brief list of sangoma translators, populated on module loading depending on which translators are available or configured */
static AST_LIST_HEAD_STATIC(g_sangoma_translators, sangoma_translator);

/* \brief list of current sangoma transcoding sessions */
static AST_LIST_HEAD_STATIC(g_sangoma_sessions, sangoma_transcoding_session);

/* \brief Module description */
static char g_sangoma_module_description[] = "Sangoma Codec Translator";

/* \brief cli handler to show sangoma transcoding usage */
#if ASTERISK_VERSION_NUM >= 10600
static char *handle_cli_sangoma_show_translators(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_cli_sangoma_show_transcoding(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_cli_sangoma_flush_transcoding(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_cli_sangoma_show_rtp_statistics(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
#else
int handle_cli_sangoma_show_translators(int fd, int argc, char *argv[]);
int handle_cli_sangoma_show_transcoding(int fd, int argc, char *argv[]);
int handle_cli_sangoma_flush_transcoding(int fd, int argc, char *argv[]);
int handle_cli_sangoma_show_rtp_statistics(int fd, int argc, char *argv[]);
#endif

static char g_cli_sangoma_show_translators_usage[] = 
			"Usage: sangoma show translators\n"
			"	Display Sangoma registered translators\n";
static char g_cli_sangoma_show_translators_description[] = "Display Sangoma registered translators.";

static char g_cli_sangoma_show_transcoding_usage[] = 
			"Usage: sangoma show transcoding sessions\n"
			"		Display Sangoma current transcoding sessions and statistics\n";
static char g_cli_sangoma_show_transcoding_description[] = "Display Sangoma transcoding utilization.";

static char g_cli_sangoma_flush_transcoding_usage[] = 
			"Usage: sangoma flush transcoding session\n"
			"		Flush all or only the provided sangoma transcoding session RTP buffers\n";
static char g_cli_sangoma_flush_transcoding_description[] = "Flush Sangoma RTP transcoding buffers (be careful).";

static char g_cli_sangoma_show_rtp_usage[] = 
			"Usage: sangoma show rtp statistics\n"
			"		Display Sangoma current RTP statistics per session\n";
static char g_cli_sangoma_show_rtp_description[] = "Display Sangoma RTP utilization.";

/* \brief array of cli entries to be registered with the Asterisk CLI core */
static struct ast_cli_entry g_sangoma_cli[] = {
#if ASTERISK_VERSION_NUM >= 10600
	AST_CLI_DEFINE(handle_cli_sangoma_show_translators, g_cli_sangoma_show_translators_description),
	AST_CLI_DEFINE(handle_cli_sangoma_show_transcoding, g_cli_sangoma_show_transcoding_description),
	AST_CLI_DEFINE(handle_cli_sangoma_flush_transcoding, g_cli_sangoma_flush_transcoding_description),
	AST_CLI_DEFINE(handle_cli_sangoma_show_rtp_statistics, g_cli_sangoma_show_rtp_description)
#else
	{
		{ "sangoma", "show", "translators", NULL },
		handle_cli_sangoma_show_translators,
		g_cli_sangoma_show_translators_description,
		g_cli_sangoma_show_translators_usage, NULL, NULL, 
	},
	{
		{ "sangoma", "show", "transcoding", "sessions", NULL },
		handle_cli_sangoma_show_transcoding,
		g_cli_sangoma_show_transcoding_description,
		g_cli_sangoma_show_transcoding_usage, NULL, NULL, 
	},
	{
		{ "sangoma", "flush", "transcoding", NULL },
		handle_cli_sangoma_flush_transcoding,
		g_cli_sangoma_flush_transcoding_description,
		g_cli_sangoma_flush_transcoding_usage, NULL, NULL, 
	},
	{
		{ "sangoma", "show", "rtp", "statistics", NULL },
		handle_cli_sangoma_show_rtp_statistics,
		g_cli_sangoma_show_rtp_description,
		g_cli_sangoma_show_rtp_usage, NULL, NULL, 
	}
	
#endif
};

#if ASTERISK_VERSION_NUM < 10400
AST_MUTEX_DEFINE_STATIC(usemutex);
static int localusecnt = 0;
#define AST_MODULE_LOAD_DECLINE -1
#define AST_MODULE_LOAD_FAILURE -1
#define AST_MODULE_LOAD_SUCCESS 0
#define ast_free free
#define ast_calloc calloc
#define ARRAY_LEN(a) sizeof(a)/sizeof(a[0])
#define ast_vasprintf vasprintf
#endif

#if ASTERISK_VERSION_NUM > 10602
#define sangoma_rtp_new(sess,localip) sess = ast_rtp_instance_new(SANGOMA_DEFAULT_RTP_STACK, g_sched, localip, NULL)
#define sangoma_rtp_write ast_rtp_instance_write
#define sangoma_rtp_read(instance) ast_rtp_instance_read(instance, 0)
#define sangoma_rtp_get_local_address(sess, bindaddr) ast_rtp_instance_get_local_address(sess, bindaddr)
#define sangoma_rtp_destroy ast_rtp_instance_destroy
#define sangoma_rtp_set_remote_address ast_rtp_instance_set_remote_address
#else
#define sangoma_rtp_new(sess,localip) sess = ast_rtp_new_with_bindaddr(g_sched, NULL, 0, 0, localip)
#define sangoma_rtp_write ast_rtp_write
#define sangoma_rtp_read ast_rtp_read
#define sangoma_rtp_get_local_address(sess, bindaddr) ast_rtp_get_us(sess, bindaddr)
#define sangoma_rtp_destroy ast_rtp_destroy
#define sangoma_rtp_set_remote_address ast_rtp_set_peer
#endif

static void sangoma_print_translators(int fd)
{
	int i = 0;
	struct sangoma_translator *cur;
	AST_LIST_LOCK(&g_sangoma_translators);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&g_sangoma_translators, cur, entry) {
#if 0
		ast_cli(fd, "%s Source Format: %s Destiny Format: %s\n", cur->t.name, ast_getformatname(1 << cur->t.srcfmt), 
				ast_getformatname(1 << cur->t.dstfmt));
#endif
		ast_cli(fd, "%s to %s\n", ast_getformatname(1 << cur->t.srcfmt), ast_getformatname(1 << cur->t.dstfmt));
		i++;
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&g_sangoma_translators);
	ast_cli(fd, "Total translators: %d\n", i);
}

static void sangoma_print_transcoding_sessions(int fd)
{
	struct sangoma_transcoding_session *cur;
	AST_LIST_LOCK(&g_sangoma_sessions);
	AST_LIST_TRAVERSE(&g_sangoma_sessions, cur, entry) {
		ast_cli(fd, "[%03llu]%s to %s\tTx Frames: %llu Rx Frames: %llu Rx Nulls %llu Rx Calls %llu AvgRxCallTime %dms AvgRxTime %dms AvgTxTime %dms\n", cur->sessid, 
				ast_getformatname(1 << cur->owner->t->srcfmt), ast_getformatname(1 << cur->owner->t->dstfmt),
				cur->txcount, cur->rxcount, cur->rxnullcount, cur->rxcallcount, cur->rxcallavgtime, cur->rxavgtime, cur->txavgtime);
	}
	ast_cli(fd, "Total Sessions: %llu\n", g_total_sessions);
	AST_LIST_UNLOCK(&g_sangoma_sessions);
}

#define BREAK_OR_CONTINUE(cursess, sessid) \
	if (sessid >= 0 && (cursess)->sessid == sessid) { \
		break; \
	} else { \
		continue; \
	}
static void sangoma_flush_transcoding_sessions(int fd, int sessid)
{
	struct sangoma_transcoding_session *cur;
	struct ast_frame *f;
	int res;

	if (sessid >= 0 ) {
		ast_cli(fd, "Flushing RTP for session %d only\n", sessid);
	} else {
		ast_log(LOG_WARNING, "Flushing all transcoding session buffers\n");
		ast_cli(fd, "Flushing RTP for all transcoding sessions\n");
	}

	AST_LIST_LOCK(&g_sangoma_sessions);
	AST_LIST_TRAVERSE(&g_sangoma_sessions, cur, entry) {

		if (sessid >= 0 && sessid != cur->sessid) {
			continue;
		}

		res = ast_mutex_lock(&cur->rxrtp->rxlock);
		if (res) {
			ast_log(LOG_ERROR, "Could not acquire rx lock during RTP flush in session %llu: (err %d)\n\n", cur->sessid, res);
			BREAK_OR_CONTINUE(cur, sessid);
		}

		do {
			f = sangoma_rtp_read(cur->rxrtp->ast_rtp);
		} while (f && f->frametype != AST_FRAME_NULL);

		cur->rxcount = 0;
		cur->rxnullcount = 0;
		cur->rxcallavgtime = 0;
		cur->rxavgtime = 0;

		ast_mutex_unlock(&cur->rxrtp->rxlock);

		res = ast_mutex_lock(&cur->txrtp->txlock);
		if (res) {
			ast_log(LOG_ERROR, "Could not acquire tx lock during RTP flush in session %llu: (err %d)\n\n", cur->sessid, res);
			BREAK_OR_CONTINUE(cur, sessid);
		}

		cur->txcount = 0;
		cur->txavgtime = 0;

		ast_mutex_unlock(&cur->txrtp->txlock);

		ast_cli(fd, "[%03llu]%s to %s\tflushed\n", cur->sessid, ast_getformatname(1 << cur->owner->t->srcfmt), ast_getformatname(1 << cur->owner->t->dstfmt));

		BREAK_OR_CONTINUE(cur, sessid);
	}
	ast_cli(fd, "Total Sessions: %llu\n", g_total_sessions);
	AST_LIST_UNLOCK(&g_sangoma_sessions);
}

static void sangoma_print_rtp_statistics(int fd)
{
	struct sangoma_transcoding_session *cur;
	AST_LIST_LOCK(&g_sangoma_sessions);
	AST_LIST_TRAVERSE(&g_sangoma_sessions, cur, entry) {
		ast_cli(fd, "[%03llu]%s to %s\tLost Frames: %li\n", cur->sessid, ast_getformatname(1 << cur->owner->t->srcfmt), 
				ast_getformatname(1 << cur->owner->t->dstfmt), cur->lostrxpackets);
	}
	AST_LIST_UNLOCK(&g_sangoma_sessions);
}

#if ASTERISK_VERSION_NUM >= 10600
static char *handle_cli_sangoma_show_translators(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "sangoma show translators";
		e->usage = g_cli_sangoma_show_translators_usage;
		break;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 3) {
		return CLI_SHOWUSAGE;
	}
	sangoma_print_translators(a->fd);
	return CLI_SUCCESS;
}

static char *handle_cli_sangoma_show_transcoding(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "sangoma show transcoding sessions";
		e->usage = g_cli_sangoma_show_transcoding_usage;
		break;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 4) {
		return CLI_SHOWUSAGE;
	}
	sangoma_print_transcoding_sessions(a->fd);
	return CLI_SUCCESS;
}

static char *handle_cli_sangoma_flush_transcoding(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int sessid = -1;
	switch (cmd) {
	case CLI_INIT:
		e->command = "sangoma flush transcoding";
		e->usage = g_cli_sangoma_flush_transcoding_usage;
		break;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc < 3) {
		return CLI_SHOWUSAGE;
	}
	if (a->argc == 4) {
		sessid = atoi(a->argv[3]);
	}
	sangoma_flush_transcoding_sessions(a->fd, sessid);
	return CLI_SUCCESS;
}

static char *handle_cli_sangoma_show_rtp_statistics(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "sangoma show rtp statistics";
		e->usage = g_cli_sangoma_show_rtp_usage;
		break;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 4) {
		return CLI_SHOWUSAGE;
	}
	sangoma_print_rtp_statistics(a->fd);
	return CLI_SUCCESS;
}
#else
int handle_cli_sangoma_show_translators(int fd, int argc, char *argv[])
{
	sangoma_print_translators(fd);
	return 0;
}

int handle_cli_sangoma_show_transcoding(int fd, int argc, char *argv[])
{
	sangoma_print_transcoding_sessions(fd);
	return 0;
}

int handle_cli_sangoma_flush_transcoding(int fd, int argc, char *argv[])
{
	int sessid = -1;
	if (argc == 4) {
		sessid = atoi(argv[3]);
	}
	sangoma_flush_transcoding_sessions(fd, sessid);
	return 0;
}

int handle_cli_sangoma_show_rtp_statistics(int fd, int argc, char *argv[])
{
	sangoma_print_rtp_statistics(fd);
	return 0;
}
#endif

static struct ast_frame *sangoma_fakesrc_sample(void)
{
	static struct ast_frame f = {
		.frametype = AST_FRAME_VOICE,
		.samples = SANGOMA_DEFAULT_SAMPLES,
		.src = __PRETTY_FUNCTION__
	};
	return &f;
}

/*! \brief enqueue input frame to sangoma session */
static int sangoma_framein(struct asterisk_transcoding_pvt *pvt, struct ast_frame *f)
{
	int difftime = 0;
	int res = 0;
	struct timeval now = {0, 0};
	struct sangoma_transcoding_session *sess = sangoma_get_pvt(pvt);

	if (!f->subclass) {
		/* we must return a fake frame for calculation purposes, 
		 * may be we should indeed send something to sangoma board? */
		sess->fakestate = FAKE_FRAME_REQUESTED;
		pvt->samples = f->samples;
		return 0;
	}
	if (f->frametype != AST_FRAME_VOICE) {
		SANG_SESS_WARN_TX("refusing to write frame of type %d to Sangoma transcoder %llu, we only accept voice in format %s\n", 
				f->frametype, sess->sessid, ast_getformatname(1 << pvt->t->srcfmt));
		return -1;
	}
	if (f->frametype != AST_FRAME_VOICE || f->subclass != (1 << pvt->t->srcfmt)) {
		SANG_SESS_WARN_TX("refusing to write voice frame in format %s to Sangoma transcoder %llu, we only accept voice in format %s\n", 
				ast_getformatname(f->subclass), sess->sessid, ast_getformatname(1 << pvt->t->srcfmt));
		return -1;
	}

	res = ast_mutex_trylock(&sess->txrtp->txlock);
	if (res) {
		ast_log(LOG_ERROR, "Could not acquire tx lock for RTP tx in session %llu: (err %d)\n", sess->sessid, res);
		return -1;
	}

	if (sangoma_rtp_write(sess->txrtp->ast_rtp, f)) {
		SANG_SESS_WARN_TX("Failed to write voice frame in format %s of length %d to Sangoma transcoder %llu\n", 
		ast_getformatname(f->subclass), f->datalen, sess->sessid);
		ast_mutex_unlock(&sess->txrtp->txlock);
		return -1;
	}

	ast_mutex_unlock(&sess->txrtp->txlock);

	now = ast_tvnow();
	if (!sess->lasttxtime.tv_sec) {
		sess->lasttxtime = now;
	} else {
		difftime = ast_tvdiff_ms(now, sess->lasttxtime);
		sess->txavgtime = sess->txavgtime ? ((sess->txavgtime + difftime) / 2) : difftime;
		sess->lasttxtime = now;
	}
	sess->txcount++;

	pvt->samples += f->samples;
	return 0;
}

/*! \brief output frames from the sangoma session */
static struct ast_frame *sangoma_frameout(struct asterisk_transcoding_pvt *pvt)
{
	/* read from the rtp session and put the result in pvt->f */
	int difftime = 0;
	int res = 0;
	struct timeval now = {0, 0};
	struct ast_frame *f = NULL;
	struct sangoma_transcoding_session *sess = sangoma_get_pvt(pvt);

	if (FAKE_FRAME_REQUESTED == sess->fakestate) {
		sess->fakestate = FAKE_FRAME_PROVIDED;
		pvt->f.frametype = AST_FRAME_VOICE;
		pvt->f.subclass = 0;
		pvt->f.samples = SANGOMA_DEFAULT_SAMPLES;
/* if any version previous to 1.6.1 defines data.ptr, then decrease this version num */
#if ASTERISK_VERSION_NUM >= 10601
		pvt->f.data.ptr = NULL;
#else
		pvt->f.data = NULL;
#endif
#if ASTERISK_VERSION_NUM >= 10400
		//ast_set_flag(&pvt->f, AST_FRFLAG_FROM_TRANSLATOR);
#endif
		pvt->f.offset = 0;
		pvt->f.datalen = 0;
		pvt->f.mallocd = 0;
		pvt->samples = 0;
		/*ast_log(LOG_DEBUG, "Returned fake frame from Sangoma transcoder\n");*/
		return ast_frisolate(&pvt->f);
		//return &pvt->f;
	} else if (FAKE_FRAME_PROVIDED == sess->fakestate) {
		/*ast_log(LOG_DEBUG, "Terminated fake frame request to Sangoma transcoder\n");*/
		sess->fakestate = FAKE_FRAME_TERMINATED;
		return NULL;
	}

	res = ast_mutex_trylock(&sess->rxrtp->rxlock);
	if (res) {
		ast_log(LOG_ERROR, "Could not acquire rx lock for RTP rx in session %llu: (err %d)\n\n", sess->sessid, res);
		return NULL;
	}

	/* if this is the first time we read we must read any previously buffered data in the RTP socket
	 * otherwise we may end up with audio delays because depending on the board configuration it is
	 * possible the board has been sending us audio even if we did not send audio */
	if (!sess->rxcount) {
		do {
			f = sangoma_rtp_read(sess->rxrtp->ast_rtp);
			sess->rxcount++;
		} while (f && f->frametype != AST_FRAME_NULL);
	} else {
		f = sangoma_rtp_read(sess->rxrtp->ast_rtp);
	}

	/* calculate average input time */
	now = ast_tvnow();
	if (!sess->lastrxcalltime.tv_sec) {
		sess->lastrxcalltime = now;
	} else {
		difftime = ast_tvdiff_ms(now, sess->lastrxcalltime);
		sess->rxcallavgtime = sess->rxcallavgtime ? ((sess->rxcallavgtime + difftime) / 2) : difftime;
		sess->lastrxcalltime = now;
	}
	sess->rxcallcount++;

	ast_mutex_unlock(&sess->rxrtp->rxlock);

	if (!f) {
		SANG_SESS_WARN_RX("ignoring read NULL frame from Sangoma transcoder %llu, we only expect voice in format %s\n", 
				sess->sessid, ast_getformatname(1 << pvt->t->dstfmt));
		return NULL;
	}

	if (f->frametype == AST_FRAME_NULL) {
		sess->rxnullcount++;
		return NULL;
	}
	if (f->frametype != AST_FRAME_VOICE) {
		SANG_SESS_WARN_RX("ignoring read frame of type %d from Sangoma transcoder %llu, we only expect voice in format %s\n", 
				f->frametype, sess->sessid, ast_getformatname(1 << pvt->t->dstfmt));
		return NULL;
	}
	if (f->subclass != (1 << pvt->t->dstfmt)) {
		SANG_SESS_WARN_RX("ignoring read voice frame in format %s from Sangoma transcoder %llu, we only expect voice in format %s rxrtp=%p\n",
				ast_getformatname(f->subclass), sess->sessid, ast_getformatname(1 << pvt->t->dstfmt), sess->rxrtp);
		return NULL;
	}

	/* calculate average read time */
	now = ast_tvnow();
	if (!sess->lastrxtime.tv_sec) {
		sess->lastrxtime = now;
	} else {
		difftime = ast_tvdiff_ms(now, sess->lastrxtime);
		sess->rxavgtime = sess->rxavgtime ? ((sess->rxavgtime + difftime) / 2) : difftime;
		sess->lastrxtime = now;
	}
	sess->rxcount++;

	/* check sequence */
	if (sess->lastrxseqno >= 0) {
		static int sessid = 0;
		if (sessid == 0) {
			sessid=sess->sessid;
		}
		if (f->seqno > (sess->lastrxseqno+2) ) {
			//ast_log(LOG_WARNING, "[%lli] Got Seq %i Expecing %li\n", sess->sessid, f->seqno, sess->lastrxseqno+1);
			sess->lostrxpackets += f->seqno - sess->lastrxseqno - 1;
		} 
	} 

	if (sess->lostrxpackets == -1) {
		sess->lostrxpackets = 0;
	}

	sess->lastrxseqno = f->seqno; 

	/* update samples */
	pvt->samples -= f->samples;
	return ast_frisolate(f);
}

static int sangoma_create_rtp_port(void *usr_priv, uint32_t host_ip, uint32_t *p_rtp_port, void **rtp_fd)
{
	struct sangoma_rtp_handle *sng_rtp;
	struct sockaddr_in localaddr;
	struct in_addr local_ip = { 0 };
	void *fd;

	local_ip.s_addr = htonl(host_ip);

	/* ask to create the session with the local IP provided */
	sangoma_rtp_new(fd, local_ip);
	if (!fd) {
		ast_log(LOG_ERROR, "Failed to create Sangoma transcoding RTP stream\n");
		return -1;
	}

	sangoma_rtp_get_local_address(fd, &localaddr);

	sng_rtp = ast_calloc(1, sizeof(*sng_rtp));
	if (!sng_rtp) {
		ast_log(LOG_ERROR, "Failed to create Sangoma transcoding RTP stream handle\n");
		sangoma_rtp_destroy(fd);
		return -1;
	}
	sng_rtp->ast_rtp = fd;
	
	ast_mutex_init(&sng_rtp->rxlock);
	ast_mutex_init(&sng_rtp->txlock);

	*p_rtp_port = ntohs(localaddr.sin_port);
	*rtp_fd=sng_rtp;
	ast_log(LOG_DEBUG, "Created Asterisk RTP %p\n", fd);
	return 0;
	
}

static int sangoma_create_rtp(void *usr_priv, sngtc_codec_request_leg_t *codec_reg_leg, sngtc_codec_reply_leg_t* codec_reply_leg, void **rtp_fd)
{
	struct sangoma_rtp_handle *sng_rtp;
	struct sockaddr_in remoteaddr;

	sng_rtp = *rtp_fd;

	/* set the socket where we will send the data to and read from */
	remoteaddr.sin_addr.s_addr = htonl(codec_reply_leg->codec_ip);
	remoteaddr.sin_port = htons(codec_reply_leg->codec_udp_port);
	sangoma_rtp_set_remote_address(sng_rtp->ast_rtp, &remoteaddr);
 
	ast_log(LOG_DEBUG, "Creating Codec on host (0x%08X/%d)  codec(0x%08X/%d) CODEC ID=%i ms=%i FD=%p \n",
				codec_reg_leg->host_ip, codec_reg_leg->host_udp_port,
				codec_reply_leg->codec_ip, codec_reply_leg->codec_udp_port, 
				codec_reg_leg->codec_id,  codec_reg_leg->ms, sng_rtp->ast_rtp);

	return 0;
}

static int sangoma_destroy_rtp(void *usr_priv, void *fd)
{
	struct sangoma_rtp_handle *rtp = fd;
	ast_log(LOG_DEBUG, "Destroying Asterisk RTP %p\n", rtp->ast_rtp);
	sangoma_rtp_destroy(rtp->ast_rtp);
	rtp->ast_rtp = NULL;
	ast_mutex_destroy(&rtp->rxlock);
	ast_mutex_destroy(&rtp->txlock);
	rtp->flags = 0;
	ast_free(rtp);
	return 0;
}

static int ast_format_to_codec_id(int astformat, int *ptime)
{
	int i;
	for (i = 0; g_codec_map[i].codec_id != -1; i++) {
		if (g_codec_map[i].astformat == astformat) {
			*ptime = g_codec_map[i].maxms;
			return g_codec_map[i].codec_id;
		}
	}
	return -1;
}

#if 0
static void print_stack()
{
	void *trace[50];
	char **symbols;
	int rc, i;

	rc = backtrace(trace, sizeof(trace)/sizeof(trace[0]));
	symbols = backtrace_symbols(trace, rc);
	for (i = 0; i < rc; i++) {
		ast_log(LOG_DEBUG, "%s\n", symbols[i]);
	}
	free(symbols);
}
#endif

/* \brief initialize the new sangoma sangoma transcoding session */
static int sangoma_new(struct asterisk_transcoding_pvt *pvt)
{
	int err = 0;
	int codec_id = -1;
	int ptime = 0;
	struct sangoma_transcoding_session *sess = sangoma_get_pvt(pvt);

	/* set a reference to our owner */
	sess->owner = pvt;
	sess->lostrxpackets = -1;
	sess->lastrxseqno = -1;

	sess->codec_request.usr_priv=sess;

	ast_log(LOG_DEBUG, "Got codec request from Asterisk core from %s (%d) to %s (%d)\n", 
			ast_getformatname(1 << pvt->t->srcfmt), pvt->t->srcfmt, ast_getformatname(1 << pvt->t->dstfmt), pvt->t->dstfmt);

	/*print_stack();*/

	codec_id = ast_format_to_codec_id((1 << pvt->t->srcfmt), &ptime);
	if (codec_id == -1) {
		ast_log(LOG_ERROR, "Failed to map source ast format %d to codec id\n", (1 << pvt->t->srcfmt));
		return -1;
	}
	sess->codec_request.a.codec_id = codec_id;
	sess->codec_request.a.ms = ptime;

	codec_id = ast_format_to_codec_id((1 << pvt->t->dstfmt), &ptime);
	if (codec_id == -1) {
		ast_log(LOG_ERROR, "Failed to map source ast format %d to codec id\n", (1 << pvt->t->dstfmt));
		return -1;
	}
	sess->codec_request.b.codec_id = codec_id;
	sess->codec_request.b.ms = ptime;


	ast_log(LOG_DEBUG, "Creating Codec A-CODEC_ID=%i  Codec B-CODEC_ID=%i\n",
				sess->codec_request.a.codec_id, sess->codec_request.b.codec_id);

	/* we re-use the global sessions lock to lock the session request in the library */
	AST_LIST_LOCK(&g_sangoma_sessions);

	/* create the actual new session */
	err = sngtc_create_transcoding_session(&sess->codec_request, &sess->codec_reply, 0);
	if (err) {
		AST_LIST_UNLOCK(&g_sangoma_sessions);
		ast_log(LOG_ERROR, "Failed to create Sangoma transcoding session\n");
		sess->rxrtp = NULL;
		sess->txrtp = NULL;
		return -1;
	}

	/* add the new session to the global list for usage statistics */
	sess->txrtp = sess->codec_reply.tx_fd;
	sess->rxrtp = sess->codec_reply.rx_fd;

	/* sanity check flags, we should not get RTP already used, if this fail is a bug in the library */
	if (ast_test_flag(sess->txrtp, SNG_RTP_TX)) {
		ast_log(LOG_ERROR, "Invalid transcoding session created! tx rtp it's already owned for transmission by someone else!\n");
		sess->rxrtp = NULL;
		sess->txrtp = NULL;
		if (sngtc_free_transcoding_session(&sess->codec_reply)) {
			ast_log(LOG_ERROR, "Failed to destroy Sangoma transcoding session properly (after discovering TX is already in use)\n");
		}
		AST_LIST_UNLOCK(&g_sangoma_sessions);
		return -1;
	}

	if (ast_test_flag(sess->rxrtp, SNG_RTP_RX)) {
		ast_log(LOG_ERROR, "Invalid transcoding session created! rx rtp it's already owned for transmission by someone else!\n");
		sess->rxrtp = NULL;
		sess->txrtp = NULL;
		if (sngtc_free_transcoding_session(&sess->codec_reply)) {
			ast_log(LOG_ERROR, "Failed to destroy Sangoma transcoding session properly (after discovering RX is already in use)\n");
		}
		AST_LIST_UNLOCK(&g_sangoma_sessions);
		return -1;
	}

	ast_set_flag(sess->txrtp, SNG_RTP_TX);
	ast_set_flag(sess->rxrtp, SNG_RTP_RX);

	sess->sessid = g_next_session_id++;
	g_total_sessions++;
	
	ast_log(LOG_DEBUG, "Created Codec ID=%llu A-CODEC_ID/IANA=%i/%i  Codec B-CODEC_ID/IANA=%i/%i  Tx=%p  Rx=%p\n",
				sess->sessid,
				sess->codec_request.a.codec_id, 
				sess->codec_reply.a.iana_code,
				sess->codec_request.b.codec_id,
				sess->codec_reply.b.iana_code,
				sess->txrtp,
				sess->rxrtp);

	AST_LIST_INSERT_HEAD(&g_sangoma_sessions, sess, entry);

	AST_LIST_UNLOCK(&g_sangoma_sessions);

	ast_log(LOG_DEBUG, "Sangoma Transcoding Session %llu Created\n", sess->sessid);

	ast_log(LOG_DEBUG, "Sangoma Transcoding Session %llu Rx %s in %d.%d.%d.%d:%d from board at %d.%d.%d.%d:%d fd=%p\n",
			sess->sessid, ast_getformatname(1 << sess->owner->t->dstfmt), 
			SNGTC_NIPV4(sess->codec_reply.b.host_ip), sess->codec_reply.b.host_udp_port,
			SNGTC_NIPV4(sess->codec_reply.b.codec_ip), sess->codec_reply.b.codec_udp_port, sess->rxrtp); 

	ast_log(LOG_DEBUG, "Sangoma Transcoding Session %llu Tx %s from %d.%d.%d.%d:%d to board at %d.%d.%d.%d:%d fd=%p\n",
			sess->sessid, ast_getformatname(1 << sess->owner->t->srcfmt),
			SNGTC_NIPV4(sess->codec_reply.a.host_ip), sess->codec_reply.a.host_udp_port,
			SNGTC_NIPV4(sess->codec_reply.a.codec_ip), sess->codec_reply.a.codec_udp_port, sess->txrtp);
	
	return 0;
}

static void sangoma_destroy(struct asterisk_transcoding_pvt *pvt)
{
	struct sangoma_transcoding_session *sess = sangoma_get_pvt(pvt);

	ast_log(LOG_DEBUG, "Destroying Sangoma transcoding session %llu\n", sess->sessid);

	AST_LIST_LOCK(&g_sangoma_sessions);
	
	/* clear sanity flags, we're no longer using this RTP session for tx/rx */
	ast_clear_flag(sess->txrtp, SNG_RTP_TX);
	ast_clear_flag(sess->rxrtp, SNG_RTP_RX);

	if (sngtc_free_transcoding_session(&sess->codec_reply)) {
		ast_log(LOG_ERROR, "Failed to destroy Sangoma transcoding session %llu properly\n", sess->sessid);
	}

	AST_LIST_REMOVE(&g_sangoma_sessions, sess, entry);
	g_total_sessions--;
	AST_LIST_UNLOCK(&g_sangoma_sessions);
}

#if ASTERISK_VERSION_NUM < 10400
#define LOCAL_COUNT_INC  \
	ast_mutex_lock(&usemutex); \
	localusecnt++; \
	ast_mutex_unlock(&usemutex); 
#define LOCAL_COUNT_DEC  \
	ast_mutex_lock(&usemutex); \
	localusecnt--; \
	ast_mutex_unlock(&usemutex); 
static struct asterisk_transcoding_pvt *sangoma_new_g729toulaw()
{
	struct asterisk_transcoding_pvt *pvt = NULL;
	LOCAL_COUNT_INC;
	pvt = calloc(1, sizeof(struct asterisk_transcoding_pvt));
	if (!pvt) {
		ast_log(LOG_ERROR, "Could not allocate pvt structure\n");
		LOCAL_COUNT_DEC;
		return NULL;
	}
	pvt->dstfmt = 2;
	pvt->srcfmt = 8;
	pvt->t = pvt;
	if (sangoma_new(pvt)) {
		free(pvt);
		LOCAL_COUNT_DEC;
		return NULL;
	}
	return pvt;
}

static struct asterisk_transcoding_pvt *sangoma_new_ulawtog729()
{
	struct asterisk_transcoding_pvt *pvt = NULL;
	LOCAL_COUNT_INC;
	pvt = calloc(1, sizeof(struct asterisk_transcoding_pvt));
	if (!pvt) {
		LOCAL_COUNT_DEC;
		ast_log(LOG_ERROR, "Could not allocate pvt structure\n");
		return NULL;
	}
	pvt->dstfmt = 8;
	pvt->srcfmt = 2;
	pvt->t = pvt;
	if (sangoma_new(pvt)) {
		free(pvt);
		LOCAL_COUNT_DEC;
		return NULL;
	}
	return pvt;
}

static void sangoma_destroy_pvt_old(struct asterisk_transcoding_pvt *pvt)
{
	sangoma_destroy(pvt);
	free(pvt);
	LOCAL_COUNT_DEC;
}
#endif /* endif ASTERISK 1.2 */

static int load_nic_network_information(char *nic, sngtc_host_nic_vocallo_cfg_t *cfg)
{
	struct ifreq ifr;
	int sock;
	char *mac;
	int j = 0, k = 0, ret = 0;
	
	sock = socket(PF_INET, SOCK_STREAM, 0);
	if (-1 == sock) {
		ast_log(LOG_ERROR, "Failed to create socket to query network configuration for NIC %s: %s\n", nic, strerror(errno));
		return -1;
	}

	strncpy(ifr.ifr_name, nic, sizeof(ifr.ifr_name)-1);
	ifr.ifr_name[sizeof(ifr.ifr_name)-1] = '\0';

	if (-1 == ioctl(sock, SIOCGIFADDR, &ifr)) {
		ast_log(LOG_ERROR, "Failed to query IP address for NIC %s: %s\n", nic, strerror(errno));
		ret = -1;
		goto done;
	}
	cfg->host_ip = ntohl(((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr.s_addr);

	if (-1 == ioctl(sock, SIOCGIFNETMASK, &ifr)) {
		ast_log(LOG_ERROR, "Failed to query network address for NIC %s: %s\n", nic, strerror(errno));
		ret = -1;
		goto done;
	}
	cfg->host_ip_netmask = ntohl(((struct sockaddr_in *)&ifr.ifr_netmask)->sin_addr.s_addr);

	if (-1 == ioctl(sock, SIOCGIFHWADDR, &ifr)) {
		ast_log(LOG_ERROR, "Failed to query HW address for NIC %s: %s\n", nic, strerror(errno));
		ret = -1;
		goto done;
	}

	mac = cfg->host_mac.mac_str;
	for (j = 0, k = 0; j < 6; j++) {
		k += snprintf(mac + k, sizeof(cfg->host_mac.mac_str) - k, j ? "-%02X" : "%02X",
				(int)(unsigned int)(unsigned char)ifr.ifr_hwaddr.sa_data[j]);
	}
	mac[sizeof(cfg->host_mac.mac_str)-1] = '\0';

done:
	close(sock);

	return ret;
}

/* updates the list of codecs and return the number of codecs/translators added */
static unsigned int update_codec_list(char codec_list[MAX_CODECS_LIST][MAX_CODEC_STR], size_t size, int index, const char *value)
{
	unsigned int i = index;
	char line[1024];
	char *c = NULL;
	char *tok = NULL;

	ast_log(LOG_DEBUG, "Parsing codec string %s\n", value);
	ast_copy_string(line, value, sizeof(line));

	/* the value string is in format codec,codec,codec where codec can either be a codec name 
	 * or a translator name (translator names are codec1tocodec2 names ) */
	tok = strtok_r(line, ",", &c);
	if (tok == NULL) {
		ast_log(LOG_WARNING, "not found any codec specification!\n");
		return 0;
	}

	ast_copy_string(codec_list[i], tok, sizeof(codec_list[0]));
	ast_log(LOG_NOTICE, "found translator %s\n", codec_list[i]);
	i++;

	while ((tok = strtok_r(NULL, ",", &c))) {
		ast_copy_string(codec_list[i], tok, sizeof(codec_list[0]));
		ast_log(LOG_NOTICE, "found translator %s\n", codec_list[i]);
		i++;
	}

	return (i - index);
}

static int sangoma_parse_config(int reload)
{
	struct ast_variable *v = NULL;
	char *category = NULL;
	int nic_idx = 0;
	char ipbuf1[255];
	char ipbuf2[255];
	const char *ipaddr = NULL;
	struct sockaddr_in vocalloaddr;
	struct in_addr print_host_ip;
	struct in_addr print_host_netmask;
#if ASTERISK_VERSION_NUM >= 10600
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };
	struct ast_config *cfg = ast_config_load(SANGOMA_TRANSCODE_CONFIG, config_flags);
	if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		ast_verbose("File " SANGOMA_TRANSCODE_CONFIG " was not changed, ignoring reload\n");
		return 0;
	}
#else
	struct ast_config *cfg = ast_config_load(SANGOMA_TRANSCODE_CONFIG);
#endif

#if ASTERISK_VERSION_NUM > 10602
	if (cfg == CONFIG_STATUS_FILEMISSING || cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_WARNING, "Sangoma transcoding configuration file " SANGOMA_TRANSCODE_CONFIG " is missing or invalid\n");
		return -1;
	}
#endif
	if (!cfg) {
		ast_log(LOG_WARNING, "Failed to Sangoma transcoding configuration file " SANGOMA_TRANSCODE_CONFIG ", make sure the file exists and is readable\n");
		return -1;
	}

	memset(&g_init_cfg, 0, sizeof(g_init_cfg));
	memset(&vocalloaddr, 0, sizeof(vocalloaddr));
	memset(g_codec_register_list, 0, sizeof(g_codec_register_list));
	memset(g_codec_noregister_list, 0, sizeof(g_codec_noregister_list));
	while ((category = ast_category_browse(cfg, category))) {

		if (!strcasecmp("general", category)) {
			for (v = ast_variable_browse(cfg, category); v; v = v->next) {
				if (!strcasecmp(v->name, "unregisterconflictingcodecs")) {
					g_attempt_unload_conflicting_translators = ast_true(v->value);
					if (g_attempt_unload_conflicting_translators) {
						ast_log(LOG_NOTICE, "unregisterconflictingcodecs is now enabled, this may cause warnings when unloading other codec modules\n");
					}
				} else if (!strcasecmp(v->name, "allownondahditiming")) {
					g_allow_nondahdi_timing = ast_true(v->value);
					if (g_allow_nondahdi_timing) {
						ast_log(LOG_WARNING, "BE AWARE THAT NON-DAHDI TIMING HAS BEEN KNOWN TO CAUSE AUDIO LOSS AND/OR CRASHES\n");
						ast_log(LOG_WARNING, "BE AWARE THAT NON-DAHDI TIMING HAS BEEN KNOWN TO CAUSE AUDIO LOSS AND/OR CRASHES\n");
						ast_log(LOG_WARNING, "BE AWARE THAT NON-DAHDI TIMING HAS BEEN KNOWN TO CAUSE AUDIO LOSS AND/OR CRASHES\n");
					}
				} else if (!strcasecmp(v->name, "register")) {
					g_codec_register_list_count += update_codec_list(g_codec_register_list, ARRAY_LEN(g_codec_register_list), g_codec_register_list_count, v->value);
				} else if (!strcasecmp(v->name, "noregister")) {
					g_codec_noregister_list_count += update_codec_list(g_codec_noregister_list, ARRAY_LEN(g_codec_noregister_list), g_codec_noregister_list_count, v->value);
				} else {
					ast_log(LOG_WARNING, "Ignored unknown Sangoma codec setting %s in [general]\n", v->name);
				}
			}
			continue;
		}

		if (load_nic_network_information(category, &g_init_cfg.host_nic_vocallo_cfg[nic_idx])) {
			ast_log(LOG_WARNING, "Ignoring NIC %s, failed to retrieve IP network configuration\n", category);
			continue;
		}
		ast_log(LOG_DEBUG, "Configuring NIC %s ( %s )\n", category, g_init_cfg.host_nic_vocallo_cfg[nic_idx].host_mac.mac_str);

		/* set defaults in case is some parameters are not specified */
		g_init_cfg.host_nic_vocallo_cfg[nic_idx].vocallo_base_udp_port = SANGOMA_DEFAULT_UDP_PORT;
		g_init_cfg.host_nic_vocallo_cfg[nic_idx].silence_suppression = 0;

		for (v = ast_variable_browse(cfg, category); v; v = v->next) {
			if (!strcasecmp(v->name, "baseudp")) {
				g_init_cfg.host_nic_vocallo_cfg[nic_idx].vocallo_base_udp_port = atoi(v->value);
				ast_log(LOG_DEBUG, "Using base udp %s (%d)\n", v->value, g_init_cfg.host_nic_vocallo_cfg[nic_idx].vocallo_base_udp_port);
			}
			else if (!strcasecmp(v->name, "vocalloaddr")) {
#if ASTERISK_VERSION_NUM >= 10600
				if (ast_parse_arg(v->value, PARSE_INADDR, &vocalloaddr)) {
					ast_log(LOG_WARNING, "Invalid Sangoma transcoding vocallo addr: %s\n", v->value);
					vocalloaddr.sin_addr.s_addr = 0;
					continue;
				}
#else
				struct ast_hostent ahp;
				struct hostent *hp;
				if (!(hp = ast_gethostbyname(v->value, &ahp))) {
					ast_log(LOG_WARNING, "Invalid Sangoma transcoding Vocallo host, could not resolve: %s\n", v->value);
					continue;
				} else {
					memcpy(&vocalloaddr.sin_addr, hp->h_addr, sizeof(vocalloaddr.sin_addr));
				}
#endif
				g_init_cfg.host_nic_vocallo_cfg[nic_idx].vocallo_ip = ntohl(vocalloaddr.sin_addr.s_addr);
				ast_log(LOG_DEBUG, "Using transcoding vocallo host %s (0x%08X)\n", v->value, 
						g_init_cfg.host_nic_vocallo_cfg[nic_idx].vocallo_ip);
			} else if (!strcasecmp(v->name, "silence")) {
				g_init_cfg.host_nic_vocallo_cfg[nic_idx].silence_suppression = ast_true(v->value) ? 1 : 0;
			} else {
				ast_log(LOG_WARNING, "Ignored unknown Sangoma codec setting %s in NIC %s configuration\n", v->name, category);
			}
		}

		/* validate some stuff */
		if (!g_init_cfg.host_nic_vocallo_cfg[nic_idx].vocallo_ip) {
			ast_log(LOG_ERROR, "Ignoring Sangoma transcoding interface %s, you must specify a valid vocalloaddr parameter\n", category);
			continue;
		}

		print_host_ip.s_addr = htonl(g_init_cfg.host_nic_vocallo_cfg[nic_idx].host_ip);
		print_host_netmask.s_addr = htonl(g_init_cfg.host_nic_vocallo_cfg[nic_idx].host_ip_netmask);
#if ASTERISK_VERSION_NUM >= 10400
		ipaddr = ast_inet_ntoa(print_host_ip);
		strncpy(ipbuf1, ipaddr, sizeof(ipbuf1)-1);
		ipbuf1[sizeof(ipbuf1)-1] = 0;

		ipaddr = ast_inet_ntoa(print_host_netmask);
		strncpy(ipbuf2, ipaddr, sizeof(ipbuf2)-1);
		ipbuf1[sizeof(ipbuf2)-1] = 0;
		ast_verbose("Sangoma transcoding interface %s using IP address %s, netmask %s, silence suppression = %s\n", 
				category, ipbuf1, ipbuf2, g_init_cfg.host_nic_vocallo_cfg[nic_idx].silence_suppression ? "yes" : "no");
#else
		ast_verbose("Sangoma transcoding interface %s using IP address %s, netmask %s\n", 
				category, ast_inet_ntoa(buf1, sizeof(ipbuf1), print_host_ip), 
				ast_inet_ntoa(buf2, sizeof(ipbuf2), print_host_netmask));
#endif
		nic_idx++;
	}

	ast_config_destroy(cfg);

	g_init_cfg.host_nic_vocallo_sz = nic_idx;

	if (!g_codec_register_list_count) {
		ast_log(LOG_NOTICE, "No 'register' codecs where specified in configuration file, defaulting to g729, g726 and g722\n");
		ast_copy_string(g_codec_register_list[0], "g729", sizeof(g_codec_register_list[0]));
		ast_copy_string(g_codec_register_list[1], "g726", sizeof(g_codec_register_list[0]));
		ast_copy_string(g_codec_register_list[2], "g722", sizeof(g_codec_register_list[0]));
		g_codec_register_list_count = 3;
	}
	return 0;
}

/*! \brief standard module stuff */

#if ASTERISK_VERSION_NUM >= 10400
static 
#endif
int reload(void)
{
	if (sangoma_parse_config(1)) {
		return AST_MODULE_LOAD_DECLINE;
	}
	return AST_MODULE_LOAD_SUCCESS;
}

static void unregister_translators(void)
{
	struct sangoma_translator *cur;
	ast_log(LOG_DEBUG, "Unregistering all translators\n");
	AST_LIST_LOCK(&g_sangoma_translators);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&g_sangoma_translators, cur, entry) {
		ast_unregister_translator(&cur->t);
		ast_free(cur);
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&g_sangoma_translators);
}

#if ASTERISK_VERSION_NUM >= 10400
static 
#endif
int unload_module(void)
{
	ast_cli_unregister_multiple(g_sangoma_cli, ARRAY_LEN(g_sangoma_cli));
	unregister_translators();
	sngtc_deactivate_modules();
	if (g_sched) {
		sched_context_destroy(g_sched);
		g_sched = NULL;
	}
	return 0;
}

static int sangoma_logger(int level, char *fmt, ...)
{
	char *data;
	int ret = 0;
	va_list ap;

	va_start(ap, fmt);
	ret = ast_vasprintf(&data, fmt, ap);
	if (ret == -1) {
		return ret;
	}
	va_end(ap);

	switch (level) {
	case SNGTC_LOGLEVEL_DEBUG:
		ast_log(LOG_DEBUG, "%s\n", data);  
		break;
	case SNGTC_LOGLEVEL_WARN:
		ast_log(LOG_WARNING, "%s\n", data);  
		break;
	case SNGTC_LOGLEVEL_INFO:
		ast_log(LOG_NOTICE,"%s\n", data);
		break;
	case SNGTC_LOGLEVEL_STATS:
		break;
	case SNGTC_LOGLEVEL_ERROR:
		ast_log(LOG_ERROR,"%s\n", data);
		break;
	case SNGTC_LOGLEVEL_CRIT:
		ast_log(LOG_ERROR,"%s\n", data);
		break;
	default:
		ast_log(LOG_ERROR,"Error: invalid loglevel %i\n",level);
		return -1;
	}

	ast_free(data);
	return 0;
}

static int register_translators(void)
{
	int ret = 0;
	struct sangoma_translator *cur;
	ast_log(LOG_DEBUG, "Registering all translators\n");
	AST_LIST_LOCK(&g_sangoma_translators);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&g_sangoma_translators, cur, entry) {
		if (ast_register_translator(&cur->t)) {
			ast_log(LOG_ERROR, "Failed to register translator %s\n", cur->t.name);
			ret = -1;
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&g_sangoma_translators);
	return ret;
}

static void attempt_unload_conflicting_translators(int srcfmt, int dstfmt)
{
	int steps = 0;
	struct ast_trans_pvt *path;
	char cname[50];
retryunload:
	steps = ast_translate_path_steps(dstfmt, srcfmt);
	if (steps == 1) {
		path = ast_translator_build_path(dstfmt, srcfmt);
		if (!path) {
			ast_log(LOG_ERROR, "Wow, steps from %d(%s) to %d(%s) but no translation path\n",
				srcfmt, ast_getformatname(srcfmt), dstfmt, ast_getformatname(dstfmt));
			return;
		}
		/* this is evil and will cause the owner modules for this translator to 
		 * fail to unregister it but is better than crashing */
		ast_copy_string(cname, path->t->name, sizeof(cname));
		if (ast_unregister_translator(path->t)) {
			ast_log(LOG_ERROR, "Failed to force unregistration of codec %s, crash may occur!", cname);
		} else {
			ast_log(LOG_NOTICE, "Forced unregistration of codec %s\n", cname);
		}
		ast_translator_free_path(path);
		goto retryunload;
	}
}

static int in_list(char codec_list[MAX_CODECS_LIST][MAX_CODEC_STR], size_t size, char *value, int explicit)
{
	int i;
	for (i = 0; i < size; i++) {
		if (!strcasecmp(value, codec_list[i])) {
			return 1;
		}
		/* if any list element has "all" and explicit search has not been requested 
		   then accept the value */
		if (!explicit && !strcasecmp(codec_list[i], "all")) {
			return 1;
		}
	}
	return 0;
}

static int print_list(char codec_list[MAX_CODECS_LIST][MAX_CODEC_STR], size_t size, const char *name)
{
	int i;
	ast_log(LOG_DEBUG, "%s list has %zd elements: \n", name, size);
	for (i = 0; i < size; i++) {
		ast_log(LOG_DEBUG, "%s\n", codec_list[i]);
	}
	return 0;
}

static int add_translator(int srcfmt, int dstfmt)
{
	struct sangoma_translator *vt;
	char tcname[255];

	snprintf(tcname, sizeof(tcname), "%sto%s", ast_getformatname(srcfmt), ast_getformatname(dstfmt));
	if (!in_list(g_codec_register_list, g_codec_register_list_count, ast_getformatname(srcfmt), 0) &&
	    !in_list(g_codec_register_list, g_codec_register_list_count, ast_getformatname(dstfmt), 0) &&
	    !in_list(g_codec_register_list, g_codec_register_list_count, tcname, 0)
	   ) {
		/* neither srcfmt, dstfmt or the full translator names were found in the register list */
		ast_log(LOG_DEBUG, "Not registering translator %s because was not found in the register list\n", tcname);
		return 0;
	}
	if (in_list(g_codec_noregister_list, g_codec_noregister_list_count, ast_getformatname(srcfmt), 1) ||
	    in_list(g_codec_noregister_list, g_codec_noregister_list_count, ast_getformatname(dstfmt), 1) ) {
		/* if the src or dest formats are in the noregister, the only way to proceeed is that 
		 * the explicit codec is in the register list */
		if (!in_list(g_codec_register_list, g_codec_register_list_count, tcname, 1)) {
			ast_log(LOG_DEBUG, "Not registering translator %s because it's src/dst was found "
					"in the noregister list and was not found explicitly in the register list\n", tcname);
			return 0;
		}
	}
		    
	if (in_list(g_codec_noregister_list, g_codec_noregister_list_count, tcname, 1)) {
		/* neither srcfmt, dstfmt or the full translator names were found in the register list */
		ast_log(LOG_DEBUG, "Not registering translator %s because it was found explicitly in the noregister list\n", tcname);
		return 0;
	}

	if (!(vt = ast_calloc(1, sizeof(*vt)))) {
		return -1;
	}

	vt->t.srcfmt = srcfmt;
	vt->t.dstfmt = dstfmt;
	vt->t.framein = sangoma_framein;
	vt->t.frameout = sangoma_frameout;
	vt->t.sample = sangoma_fakesrc_sample;
#if ASTERISK_VERSION_NUM >= 10400
	vt->t.desc_size = sizeof(struct sangoma_transcoding_session);
	vt->t.newpvt = sangoma_new;
	vt->t.destroy = sangoma_destroy;
	vt->t.buf_size = SANGOMA_TRANSLATOR_BUFFER_SIZE;
#else
#error "FIXME: Asterisk 1.2 not supported yet"
	vt->t.newpvt = i == 0 ? sangoma_new_g729toulaw : sangoma_new_ulawtog729;
	vt->t.destroy = sangoma_destroy_pvt_old;
#endif
	snprintf((char *)vt->t.name, sizeof(vt->t.name), "sng%s", tcname);

	if (g_attempt_unload_conflicting_translators) {
		attempt_unload_conflicting_translators(srcfmt, dstfmt);
	}

	/* it is important to NOT register the translator just yet, just add it to our linked list 
	 * reason being that attempt_unload_conflicting_translators() has better chances to succed
	 * if we dont register our translations in the mix */
	AST_LIST_LOCK(&g_sangoma_translators);
	AST_LIST_INSERT_HEAD(&g_sangoma_translators, vt, entry);
	AST_LIST_UNLOCK(&g_sangoma_translators);

	return 0;
}

static int entry_callback(const char *module, const char *desc, int usecnt, const char *like)
{
	if (strcasestr(module, "res_timing_dahdi")) {
		g_dahdi_timing = 1;
		return 1;
	}

	if (strcasestr(module, like)) {
		return 1;
	}

	return 0;
}

static int check_dahdi_timing(void)
{
#if ASTERISK_VERSION_NUM > 10602
	int timing_sources = 0;
	if (g_allow_nondahdi_timing) {
		return 0;
	}
	timing_sources = ast_update_module_list(entry_callback, "res_timing_");
	if (timing_sources != 1 || !g_dahdi_timing) {
		return -1;
	}
#endif
	if (0) {
		entry_callback(NULL, NULL, 0, NULL);
	}
	return 0;
}

#if ASTERISK_VERSION_NUM >= 10400
static 
#endif
int load_module(void)
{
	struct in_addr localhost_ip = { 0 };
	void *fakertp = NULL;
	int i, j, astiana;
	int detected = 0, activated = 0;
	uid_t procuid;

	/* this module can only be run as root */
	procuid = geteuid();
	if (procuid != 0) {
		ast_log(LOG_ERROR, "This module requires to create raw sockets and only root can do that. Please run Asterisk as root\n");
		return AST_MODULE_LOAD_FAILURE;
	}

	if (sangoma_parse_config(0)) {
		return AST_MODULE_LOAD_FAILURE;
	}

	if (check_dahdi_timing()) {
		ast_log(LOG_ERROR, "Refusing to load this module with non-dahdi timing, please disable all res_timing_*.so modules and leave res_timing_dahdi.so as the only timing source in modules.conf\n");
		ast_log(LOG_ERROR, "You can load this module with other timing sources by setting allownondahditiming=yes in %s general configuration section\n", SANGOMA_TRANSCODE_CONFIG);
		return AST_MODULE_LOAD_FAILURE;
	}

	g_init_cfg.log = sangoma_logger;
	g_init_cfg.create_rtp = sangoma_create_rtp;
	g_init_cfg.create_rtp_port = sangoma_create_rtp_port;
	g_init_cfg.destroy_rtp = sangoma_destroy_rtp;

	if (!(g_sched = sched_context_create())) {
		ast_log(LOG_ERROR, "Unable to create scheduler context\n");
		return AST_MODULE_LOAD_FAILURE;
	}

	sngtc_deactivate_modules();

	if (sngtc_detect_init_modules(&g_init_cfg, &detected)) {
		ast_log(LOG_ERROR, "Failed to detect init sangoma transcoding library\n");
		goto failure;
	}
    
	if (sngtc_activate_modules(&g_init_cfg, &activated)) {
		ast_log(LOG_ERROR, "Failed to activate sangoma transcoding library\n");
		goto failure;
	}   

	/* call sngtc_set_iana_code_based_on_codec_id to set Asterisk IANA codes (see main/rtp.c) */
	inet_aton("127.0.0.1", &localhost_ip);

	/* ask to create the session with the local IP provided */
	sangoma_rtp_new(fakertp, localhost_ip);
	if (!fakertp) {
		ast_log(LOG_ERROR, "Failed to create fake RTP session on 127.0.0.1 on startup\n");
		goto failure;
	}

	/* add all supported translators */
	ast_log(LOG_DEBUG, "Registering translators\n");
	print_list(g_codec_register_list, g_codec_register_list_count, "register");
	print_list(g_codec_noregister_list, g_codec_noregister_list_count, "noregister");
	for (i = 0; g_codec_map[i].codec_id != -1; i++) {

		astiana = ast_rtp_lookup_code(fakertp, 1, g_codec_map[i].astformat);
		ast_log(LOG_DEBUG, "Mapping %s to IANA %d\n", ast_getformatname(g_codec_map[i].astformat), astiana);
		sngtc_set_iana_code_based_on_codec_id(g_codec_map[i].codec_id, astiana);

		for (j = i; g_codec_map[j].codec_id != -1; j++) {

			if (g_codec_map[i].codec_id == g_codec_map[j].codec_id) {
				continue;
			}

			if (add_translator(g_codec_map[i].astformat, g_codec_map[j].astformat)) {
				goto failure;
			}

			if (add_translator(g_codec_map[j].astformat, g_codec_map[i].astformat)) {
				goto failure;
			}
		}
	}

	sangoma_rtp_destroy(fakertp);

	/* register all translators with asterisk */
	if (register_translators()) {
		goto failure;
	}

	/* register CLI helpers */
	ast_cli_register_multiple(g_sangoma_cli, ARRAY_LEN(g_sangoma_cli));

	return AST_MODULE_LOAD_SUCCESS;

failure:
	unregister_translators();
	sched_context_destroy(g_sched);
	g_sched = NULL;
	return AST_MODULE_LOAD_FAILURE;
}

#if ASTERISK_VERSION_NUM >= 10400
AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_DEFAULT, g_sangoma_module_description,
		.load = load_module,
		.unload = unload_module,
		.reload = reload,
);
#else
char *description(void)
{
	return g_sangoma_module_description;
}

int usecount(void)
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
#endif

/* For Emacs:
 * Local Variables:
 * mode:c
 * indent-tabs-mode:t
 * tab-width:4
 * c-basic-offset:4
 * End:
 * For VIM:
 * vim:set softtabstop=4 shiftwidth=4 tabstop=4
 */
